#include <xmms/plugin.h>
#include <xmms/configfile.h>
#include <xmms/util.h>

#include "squaretone-xmms.h"
#define OUT_BUFSIZE 4096

gboolean going = FALSE;
gboolean audio_error = FALSE;
pthread_t play_thread;

int seek_to;

const int OUT_FREQ = 44100;
const int BYTES_PER_SAMPLE = 4;	// 16 bit stereo


static char OutBuffer[OUT_BUFSIZE];
char *Buffer1, *Buffer2, *PBuffer;

//SQTFile sqt_file;
int samples_per_regs;
int end;

int BufLen_ms = 300;
int NumberOfBuffers = 2;
int NumberOfChannels = 2;
int SampleBit = 16;
int SampleRate = 44100;
int BlockAlign;
int BufferLength;
int BuffLen;

// For export
static InputPlugin sqt_ip = {
  NULL,
  NULL,
  NULL,				/* Description */
  sqt_init,				// sqt_init,
  sqt_about,
  NULL,
  sqt_is_our_file,
  NULL,
  sqt_play_file,
  sqt_stop,
  sqt_pause,
  sqt_seek,
  NULL,
  sqt_get_time,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  sqt_get_song_info,
  NULL,				/* file_info_box */
  NULL
};

extern void squaretone_open(char* filename);
extern void squaretone_play();
extern void squaretone_stop();
extern void squaretone_init();
extern int squaretone_make_buffer(char* buf);
extern int squaretone_get_time();
// called from xmms for plug
InputPlugin *
get_iplugin_info (void)
{
  sqt_ip.description = g_strdup_printf (_("SquareTone player (%s)"), "AyEmul2.7b3");
  return &sqt_ip;
}

void sqt_init(void)
{
  squaretone_init();
}
void
sqt_about (void)
{
}


/* Return 1 if the plugin can handle the file */
int
sqt_is_our_file (char *filename)
{
  return squaretone_is_our_file(filename);
}

int makeMix(char* stream, int len, int pos) {
  int partial;
  //   if Real_End then exit;
  
  //This code is needed due to different sizes
  //between the YM-buffer(32680) and SDLs mixing
  //buffer(4096, 32768 etc... 2^x).
  if (pos + len > BuffLen * BlockAlign) {
    //SDL wants to mix len nr of samples,
    //but our YM-buffer dont hold that much data,
    //(and not filling the complete SDL stream
    //will produce clicks),
    //so mix whats in the buffer then generate
    //another buffer and mix the missing part
    //from there.
      
    //mix whats available
    partial = BuffLen * BlockAlign - pos;
    memcpy(stream, &PBuffer[pos],
	   partial);
    //sqt_ip.output->write_audio (&PBuffer[pos], partial);

    //swap buffers
    if(PBuffer == Buffer1) 
      PBuffer = Buffer2;
    else 
      PBuffer = Buffer1;
    

    BuffLen = squaretone_make_buffer(PBuffer);
    
    //mix the rest
    stream += partial;
    memcpy(stream, &PBuffer[0],
	   len-partial);
    //sqt_ip.output->write_audio (&PBuffer[0], len-partial);
    
    //set correct pos for next mix...
    pos = len-partial;
    } else {
      memcpy(stream, &PBuffer[pos],
	     len);
      //sqt_ip.output->write_audio (&PBuffer[pos], len);
      
      pos = pos + len;
    }

  return pos;
}

void *
play_loop (void *args)
{
  int pos = 0;
  int seek_pos;
  while (going && squaretone_real_end()!=1) {

    if (seek_pos = seek_to, seek_pos != -1) {
      seek_to = -1;
      //seek_pos = api68_seek(sc68, seek_pos);
      if (seek_pos != -1) {
	//api68_debug("play_loop : flush ouput to %d\n", seek_pos);
	sqt_ip.output->flush (seek_pos);
      }
    }

    pos = makeMix(OutBuffer, OUT_BUFSIZE, pos);

    sqt_ip.add_vis_pcm(sqt_ip.output->written_time(),
		       FMT_S16_LE, 2, OUT_BUFSIZE, OutBuffer);
    while (sqt_ip.output->buffer_free() < OUT_BUFSIZE && going) {
      xmms_usleep(30000);
    }
    if (going) {
      sqt_ip.output->write_audio(OutBuffer, OUT_BUFSIZE);
    }
    if (going == 2) {
      going = 1;
    }
  }
  going = 0;
  /*FIXME: CHECK AUDIO OPEN?!*/
  sqt_ip.output->buffer_free();
  sqt_ip.output->buffer_free();

  pthread_exit (NULL);
}

int roundPow2(float n) {
  if ( n > 0 )
    return round ( exp( trunc( log(n)/log(2) ) * log(2) ));
  else //FIXME throw Error?
    return 0;
}


void
sqt_play_file (char *filename)
{
  squaretone_open(filename);
  squaretone_play();

  //YM-buffer rounded to a power of two for efficiency (not requirement)
  BufferLength = roundPow2(BufLen_ms * SampleRate / 1000);
  BlockAlign = (SampleBit / 8) * NumberOfChannels;
  Buffer1 = malloc(65535);
  Buffer2 = malloc(65535);

  PBuffer = Buffer1;
  BuffLen = squaretone_make_buffer(&PBuffer[0]);

  if (sqt_ip.output->open_audio (FMT_S16_NE, SampleRate, NumberOfChannels) == 0)
    {
      fprintf (stderr, "libsqt: output audio error!\n");
      audio_error = TRUE;
      going = FALSE;
      return;
    }
    
  sqt_ip.set_info(NULL, squaretone_get_time(), 0, SampleRate, NumberOfChannels);
  
  going = TRUE;
  pthread_create (&play_thread, NULL, play_loop, NULL);
}



void
sqt_stop (void)
{
  pthread_t p;

  going = 0;
  if (p = play_thread, p) {
    pthread_join(p, NULL);
    sqt_ip.output->close_audio ();
    play_thread = 0;
  }
  squaretone_stop();
  free(Buffer1);
  free(Buffer2);
}



void
sqt_seek (int time)
{
  xmms_usleep (10000);
}



/* Pause or unpause */
void
sqt_pause (short p)
{
  //  sqt_ip.output->pause (p);
}



/* Get the time, usually returns the output plugins output time */
int
sqt_get_time (void)
{
  int t;

  if (!going && !sqt_ip.output->buffer_playing()) {
    t = -1;
  } else {
    t = sqt_ip.output->output_time();
  }

  return t;
}


/* Function to grab the title string */
void
sqt_get_song_info (char *filename, char **title, int *length)
{
  (*length) = -1;
  (*title) = 0;
}
